@@ -14,9 +14,9 @@
  * limitations under the License.
  */
 
-#include <ctype.h>
 #include <string.h>
 
+#define CFP_CFISH
 #define C_CFISH_OBJ
 #define C_CFISH_CLASS
 #define C_CFISH_FLOAT
@@ -30,27 +30,37 @@
 #include "Clownfish/HashIterator.h"
 #include "Clownfish/Method.h"
 #include "Clownfish/Num.h"
+#include "Clownfish/PtrHash.h"
 #include "Clownfish/TestHarness/TestUtils.h"
 #include "Clownfish/Util/Atomic.h"
-#include "Clownfish/Util/StringHelper.h"
 #include "Clownfish/Util/Memory.h"
 
 #define XSBIND_REFCOUNT_FLAG   1
 #define XSBIND_REFCOUNT_SHIFT  1
 
+// Used to remember converted objects in array and hash conversion to
+// handle circular references. The root object and SV are stored separately
+// to allow lazy creation of the seen PtrHash.
+typedef struct {
+    cfish_Obj     *root_obj;
+    SV            *root_sv;
+    cfish_PtrHash *seen;
+} cfish_ConversionCache;
+
 static bool
 S_maybe_perl_to_cfish(pTHX_ SV *sv, cfish_Class *klass, bool increment,
-                      void *allocation, cfish_Obj **obj_ptr);
+                      void *allocation, cfish_ConversionCache *cache,
+                      cfish_Obj **obj_ptr);
 
 // Convert a Perl hash into a Clownfish Hash.  Caller takes responsibility for
 // a refcount.
 static cfish_Hash*
-S_perl_hash_to_cfish_hash(pTHX_ HV *phash);
+S_perl_hash_to_cfish_hash(pTHX_ HV *phash, cfish_ConversionCache *cache);
 
 // Convert a Perl array into a Clownfish Vector.  Caller takes responsibility
 // for a refcount.
 static cfish_Vector*
-S_perl_array_to_cfish_array(pTHX_ AV *parray);
+S_perl_array_to_cfish_array(pTHX_ AV *parray, cfish_ConversionCache *cache);
 
 cfish_Obj*
 XSBind_new_blank_obj(pTHX_ SV *either_sv) {
@@ -103,7 +113,7 @@ XSBind_sv_true(pTHX_ SV *sv) {
 cfish_Obj*
 XSBind_perl_to_cfish(pTHX_ SV *sv, cfish_Class *klass) {
     cfish_Obj *retval = NULL;
-    if (!S_maybe_perl_to_cfish(aTHX_ sv, klass, true, NULL, &retval)) {
+    if (!S_maybe_perl_to_cfish(aTHX_ sv, klass, true, NULL, NULL, &retval)) {
         THROW(CFISH_ERR, "Can't convert to %o", CFISH_Class_Get_Name(klass));
     }
     else if (!retval) {
@@ -115,7 +125,7 @@ XSBind_perl_to_cfish(pTHX_ SV *sv, cfish_Class *klass) {
 cfish_Obj*
 XSBind_perl_to_cfish_nullable(pTHX_ SV *sv, cfish_Class *klass) {
     cfish_Obj *retval = NULL;
-    if (!S_maybe_perl_to_cfish(aTHX_ sv, klass, true, NULL, &retval)) {
+    if (!S_maybe_perl_to_cfish(aTHX_ sv, klass, true, NULL, NULL, &retval)) {
         THROW(CFISH_ERR, "Can't convert to %o", CFISH_Class_Get_Name(klass));
     }
     return retval;
@@ -124,7 +134,9 @@ XSBind_perl_to_cfish_nullable(pTHX_ SV *sv, cfish_Class *klass) {
 cfish_Obj*
 XSBind_perl_to_cfish_noinc(pTHX_ SV *sv, cfish_Class *klass, void *allocation) {
     cfish_Obj *retval = NULL;
-    if (!S_maybe_perl_to_cfish(aTHX_ sv, klass, false, allocation, &retval)) {
+    if (!S_maybe_perl_to_cfish(aTHX_ sv, klass, false, allocation, NULL,
+                               &retval)
+       ) {
         THROW(CFISH_ERR, "Can't convert to %o", CFISH_Class_Get_Name(klass));
     }
     else if (!retval) {
@@ -135,7 +147,8 @@ XSBind_perl_to_cfish_noinc(pTHX_ SV *sv, cfish_Class *klass, void *allocation) {
 
 static bool
 S_maybe_perl_to_cfish(pTHX_ SV *sv, cfish_Class *klass, bool increment,
-                      void *allocation, cfish_Obj **obj_ptr) {
+                      void *allocation, cfish_ConversionCache *cache,
+                      cfish_Obj **obj_ptr) {
     if (sv_isobject(sv)) {
         cfish_String *class_name = CFISH_Class_Get_Name(klass);
         // Assume that the class name is always NULL-terminated. Somewhat
@@ -161,13 +174,13 @@ S_maybe_perl_to_cfish(pTHX_ SV *sv, cfish_Class *klass, bool increment,
         if (inner_type == SVt_PVAV) {
             if (klass == CFISH_VECTOR || klass == CFISH_OBJ) {
                 obj = (cfish_Obj*)
-                         S_perl_array_to_cfish_array(aTHX_ (AV*)inner);
+                         S_perl_array_to_cfish_array(aTHX_ (AV*)inner, cache);
             }
         }
         else if (inner_type == SVt_PVHV) {
             if (klass == CFISH_HASH || klass == CFISH_OBJ) {
                 obj = (cfish_Obj*)
-                         S_perl_hash_to_cfish_hash(aTHX_ (HV*)inner);
+                         S_perl_hash_to_cfish_hash(aTHX_ (HV*)inner, cache);
             }
         }
         else if (inner_type < SVt_PVAV && !SvOK(inner)) {
@@ -250,10 +263,40 @@ XSBind_hash_key_to_utf8(pTHX_ HE *entry, STRLEN *size_ptr) {
 }
 
 static cfish_Hash*
-S_perl_hash_to_cfish_hash(pTHX_ HV *phash) {
+S_perl_hash_to_cfish_hash(pTHX_ HV *phash, cfish_ConversionCache *cache) {
+    cfish_ConversionCache new_cache;
+
+    if (cache) {
+        // Lookup perl hash in conversion cache.
+        if ((SV*)phash == cache->root_sv) {
+            return (cfish_Hash*)CFISH_INCREF(cache->root_obj);
+        }
+        if (cache->seen) {
+            void *cached_hash = CFISH_PtrHash_Fetch(cache->seen, phash);
+            if (cached_hash) {
+                return (cfish_Hash*)CFISH_INCREF(cached_hash);
+            }
+        }
+    }
+
     uint32_t    num_keys = hv_iterinit(phash);
     cfish_Hash *retval   = cfish_Hash_new(num_keys);
 
+    if (!cache) {
+        // Set up conversion cache.
+        cache = &new_cache;
+        cache->root_obj = (cfish_Obj*)retval;
+        cache->root_sv  = (SV*)phash;
+        cache->seen     = NULL;
+    }
+    else {
+        if (!cache->seen) {
+            // Create PtrHash lazily.
+            cache->seen = cfish_PtrHash_new(0);
+        }
+        CFISH_PtrHash_Store(cache->seen, phash, retval);
+    }
+
     while (num_keys--) {
         HE         *entry    = hv_iternext(phash);
         STRLEN      key_len  = 0;
@@ -261,31 +304,77 @@ S_perl_hash_to_cfish_hash(pTHX_ HV *phash) {
         SV         *value_sv = HeVAL(entry);
 
         // Recurse.
-        cfish_Obj *value
-            = XSBind_perl_to_cfish_nullable(aTHX_ value_sv, CFISH_OBJ);
+        cfish_Obj *value;
+        bool success = S_maybe_perl_to_cfish(aTHX_ value_sv, CFISH_OBJ,
+                                             true, NULL, cache, &value);
+        if (!success) {
+            THROW(CFISH_ERR, "Can't convert to Clownfish::Obj");
+        }
 
         CFISH_Hash_Store_Utf8(retval, key_str, key_len, value);
     }
 
+    if (cache == &new_cache && cache->seen) {
+        CFISH_PtrHash_Destroy(cache->seen);
+    }
+
     return retval;
 }
 
 static cfish_Vector*
-S_perl_array_to_cfish_array(pTHX_ AV *parray) {
+S_perl_array_to_cfish_array(pTHX_ AV *parray, cfish_ConversionCache *cache) {
+    cfish_ConversionCache new_cache;
+
+    if (cache) {
+        // Lookup perl array in conversion cache.
+        if ((SV*)parray == cache->root_sv) {
+            return (cfish_Vector*)CFISH_INCREF(cache->root_obj);
+        }
+        if (cache->seen) {
+            void *cached_vector = CFISH_PtrHash_Fetch(cache->seen, parray);
+            if (cached_vector) {
+                return (cfish_Vector*)CFISH_INCREF(cached_vector);
+            }
+        }
+    }
+
     const uint32_t  size   = av_len(parray) + 1;
     cfish_Vector   *retval = cfish_Vec_new(size);
 
+    if (!cache) {
+        // Set up conversion cache.
+        cache = &new_cache;
+        cache->root_obj = (cfish_Obj*)retval;
+        cache->root_sv  = (SV*)parray;
+        cache->seen     = NULL;
+    }
+    else {
+        if (!cache->seen) {
+            // Create PtrHash lazily.
+            cache->seen = cfish_PtrHash_new(0);
+        }
+        CFISH_PtrHash_Store(cache->seen, parray, retval);
+    }
+
     // Iterate over array elems.
     for (uint32_t i = 0; i < size; i++) {
         SV **elem_sv = av_fetch(parray, i, false);
         if (elem_sv) {
-            cfish_Obj *elem
-                = XSBind_perl_to_cfish_nullable(aTHX_ *elem_sv, CFISH_OBJ);
+            cfish_Obj *elem;
+            bool success = S_maybe_perl_to_cfish(aTHX_ *elem_sv, CFISH_OBJ,
+                                                 true, NULL, cache, &elem);
+            if (!success) {
+                THROW(CFISH_ERR, "Can't convert to Clownfish::Obj");
+            }
             if (elem) { CFISH_Vec_Store(retval, i, elem); }
         }
     }
     CFISH_Vec_Resize(retval, size); // needed if last elem is NULL
 
+    if (cache == &new_cache && cache->seen) {
+        CFISH_PtrHash_Destroy(cache->seen);
+    }
+
     return retval;
 }
 
@@ -391,7 +480,9 @@ XSBind_arg_to_cfish(pTHX_ SV *value, const char *label, cfish_Class *klass,
                     void *allocation) {
     cfish_Obj *obj = NULL;
 
-    if (!S_maybe_perl_to_cfish(aTHX_ value, klass, false, allocation, &obj)) {
+    if (!S_maybe_perl_to_cfish(aTHX_ value, klass, false, allocation, NULL,
+                               &obj)
+       ) {
         THROW(CFISH_ERR, "Invalid value for '%s' - not a %o", label,
               CFISH_Class_Get_Name(klass));
         CFISH_UNREACHABLE_RETURN(cfish_Obj*);
@@ -409,7 +500,9 @@ XSBind_arg_to_cfish_nullable(pTHX_ SV *value, const char *label,
                              cfish_Class *klass, void *allocation) {
     cfish_Obj *obj = NULL;
 
-    if (!S_maybe_perl_to_cfish(aTHX_ value, klass, false, allocation, &obj)) {
+    if (!S_maybe_perl_to_cfish(aTHX_ value, klass, false, allocation, NULL,
+                               &obj)
+       ) {
         THROW(CFISH_ERR, "Invalid value for '%s' - not a %o", label,
               CFISH_Class_Get_Name(klass));
         CFISH_UNREACHABLE_RETURN(cfish_Obj*);
@@ -428,6 +521,37 @@ XSBind_undef_arg_error(pTHX_ const char *label) {
     THROW(CFISH_ERR, "'%s' must not be undef", label);
 }
 
+void
+XSBind_bootstrap(pTHX_ size_t num_classes,
+                 const XSBind_ClassSpec *class_specs,
+                 const XSBind_XSubSpec *xsub_specs,
+                 const char *file) {
+    size_t xsub_idx = 0;
+
+    for (size_t i = 0; i < num_classes; i++) {
+        const XSBind_ClassSpec *class_spec = &class_specs[i];
+
+        // Set up @ISA array.
+        if (class_spec->parent_name) {
+            cfish_String *isa_name
+                = cfish_Str_newf("%s::ISA", class_spec->name);
+            AV *isa = get_av(CFISH_Str_Get_Ptr8(isa_name), 1);
+            av_push(isa, newSVpv(class_spec->parent_name, 0));
+            CFISH_DECREF(isa_name);
+        }
+
+        // Register XSUBs.
+        for (uint32_t j = 0; j < class_spec->num_methods; j++) {
+            const XSBind_XSubSpec *xsub_spec = &xsub_specs[xsub_idx++];
+
+            cfish_String *xsub_name
+                = cfish_Str_newf("%s::%s", class_spec->name, xsub_spec->alias);
+            newXS(CFISH_Str_Get_Ptr8(xsub_name), xsub_spec->xsub, file);
+            CFISH_DECREF(xsub_name);
+        }
+    }
+}
+
 /***************************************************************************
  * The routines below are declared within the Clownfish core but left
  * unimplemented and must be defined for each host language.
@@ -625,7 +749,8 @@ XSBind_cfish_obj_to_sv_noinc(pTHX_ cfish_Obj *obj) {
 }
 
 void*
-CFISH_Obj_To_Host_IMP(cfish_Obj *self) {
+CFISH_Obj_To_Host_IMP(cfish_Obj *self, void *vcache) {
+    CFISH_UNUSED_VAR(vcache);
     dTHX;
     return XSBind_cfish_obj_to_sv_inc(aTHX_ self);
 }
@@ -658,8 +783,8 @@ cfish_Class_register_with_host(cfish_Class *singleton, cfish_Class *parent) {
     SAVETMPS;
     EXTEND(SP, 2);
     PUSHMARK(SP);
-    mPUSHs((SV*)CFISH_Class_To_Host(singleton));
-    mPUSHs((SV*)CFISH_Class_To_Host(parent));
+    mPUSHs((SV*)CFISH_Class_To_Host(singleton, NULL));
+    mPUSHs((SV*)CFISH_Class_To_Host(parent, NULL));
     PUTBACK;
     call_pv("Clownfish::Class::_register", G_VOID | G_DISCARD);
     FREETMPS;
@@ -674,7 +799,7 @@ cfish_Class_fresh_host_methods(cfish_String *class_name) {
     SAVETMPS;
     EXTEND(SP, 1);
     PUSHMARK(SP);
-    mPUSHs((SV*)CFISH_Str_To_Host(class_name));
+    mPUSHs((SV*)CFISH_Str_To_Host(class_name, NULL));
     PUTBACK;
     call_pv("Clownfish::Class::_fresh_host_methods", G_SCALAR);
     SPAGAIN;
@@ -694,7 +819,7 @@ cfish_Class_find_parent_class(cfish_String *class_name) {
     SAVETMPS;
     EXTEND(SP, 1);
     PUSHMARK(SP);
-    mPUSHs((SV*)CFISH_Str_To_Host(class_name));
+    mPUSHs((SV*)CFISH_Str_To_Host(class_name, NULL));
     PUTBACK;
     call_pv("Clownfish::Class::_find_parent_class", G_SCALAR);
     SPAGAIN;
@@ -738,7 +863,10 @@ void
 cfish_Err_init_class(void) {
     dTHX;
     char *file = (char*)__FILE__;
-    attempt_xsub = (SV*)newXS(NULL, cfish_Err_attempt_via_xs, file);
+    SV *xsub = (SV*)newXS(NULL, cfish_Err_attempt_via_xs, file);
+    if (!cfish_Atomic_cas_ptr((void**)&attempt_xsub, NULL, xsub)) {
+        SvREFCNT_dec(xsub);
+    }
 }
 
 cfish_Err*
@@ -769,7 +897,7 @@ cfish_Err_set_error(cfish_Err *error) {
     PUSHMARK(SP);
     PUSHmortal;
     if (error) {
-        mPUSHs((SV*)CFISH_Err_To_Host(error));
+        mPUSHs((SV*)CFISH_Err_To_Host(error, NULL));
     }
     else {
         PUSHmortal;
@@ -784,7 +912,7 @@ void
 cfish_Err_do_throw(cfish_Err *err) {
     dTHX;
     dSP;
-    SV *error_sv = (SV*)CFISH_Err_To_Host(err);
+    SV *error_sv = (SV*)CFISH_Err_To_Host(err, NULL);
     CFISH_DECREF(err);
     ENTER;
     SAVETMPS;
@@ -806,7 +934,7 @@ cfish_Err_throw_mess(cfish_Class *klass, cfish_String *message) {
 void
 cfish_Err_warn_mess(cfish_String *message) {
     dTHX;
-    SV *error_sv = (SV*)CFISH_Str_To_Host(message);
+    SV *error_sv = (SV*)CFISH_Str_To_Host(message, NULL);
     CFISH_DECREF(message);
     warn("%s", SvPV_nolen(error_sv));
     SvREFCNT_dec(error_sv);
@@ -861,7 +989,8 @@ cfish_Err_trap(CFISH_Err_Attempt_t routine, void *context) {
 /**************************** Clownfish::String *****************************/
 
 void*
-CFISH_Str_To_Host_IMP(cfish_String *self) {
+CFISH_Str_To_Host_IMP(cfish_String *self, void *vcache) {
+    CFISH_UNUSED_VAR(vcache);
     dTHX;
     SV *sv = newSVpvn(CFISH_Str_Get_Ptr8(self), CFISH_Str_Get_Size(self));
     SvUTF8_on(sv);
@@ -871,7 +1000,8 @@ CFISH_Str_To_Host_IMP(cfish_String *self) {
 /***************************** Clownfish::Blob ******************************/
 
 void*
-CFISH_Blob_To_Host_IMP(cfish_Blob *self) {
+CFISH_Blob_To_Host_IMP(cfish_Blob *self, void *vcache) {
+    CFISH_UNUSED_VAR(vcache);
     dTHX;
     return newSVpvn(CFISH_Blob_Get_Buf(self), CFISH_Blob_Get_Size(self));
 }
@@ -879,7 +1009,8 @@ CFISH_Blob_To_Host_IMP(cfish_Blob *self) {
 /**************************** Clownfish::ByteBuf ****************************/
 
 void*
-CFISH_BB_To_Host_IMP(cfish_ByteBuf *self) {
+CFISH_BB_To_Host_IMP(cfish_ByteBuf *self, void *vcache) {
+    CFISH_UNUSED_VAR(vcache);
     dTHX;
     return newSVpvn(CFISH_BB_Get_Buf(self), CFISH_BB_Get_Size(self));
 }
@@ -887,36 +1018,107 @@ CFISH_BB_To_Host_IMP(cfish_ByteBuf *self) {
 /**************************** Clownfish::Vector *****************************/
 
 void*
-CFISH_Vec_To_Host_IMP(cfish_Vector *self) {
+CFISH_Vec_To_Host_IMP(cfish_Vector *self, void *vcache) {
     dTHX;
+    cfish_ConversionCache *cache = (cfish_ConversionCache*)vcache;
+    cfish_ConversionCache  new_cache;
+
+    if (cache) {
+        // Lookup Vector in conversion cache.
+        if ((cfish_Obj*)self == cache->root_obj) {
+            return newRV_inc(cache->root_sv);
+        }
+        if (cache->seen) {
+            void *cached_av = CFISH_PtrHash_Fetch(cache->seen, self);
+            if (cached_av) {
+                return newRV_inc((SV*)cached_av);
+            }
+        }
+    }
+
     AV *perl_array = newAV();
-    uint32_t num_elems = CFISH_Vec_Get_Size(self);
+
+    if (!cache) {
+        // Set up conversion cache.
+        cache = &new_cache;
+        cache->root_obj = (cfish_Obj*)self;
+        cache->root_sv  = (SV*)perl_array;
+        cache->seen     = NULL;
+    }
+    else {
+        if (!cache->seen) {
+            // Create PtrHash lazily.
+            cache->seen = cfish_PtrHash_new(0);
+        }
+        CFISH_PtrHash_Store(cache->seen, self, perl_array);
+    }
+
+    size_t num_elems = CFISH_Vec_Get_Size(self);
 
     // Iterate over array elems.
     if (num_elems) {
+        if (num_elems > I32_MAX) {
+            THROW(CFISH_ERR, "Vector too large for Perl AV");
+        }
         av_fill(perl_array, num_elems - 1);
-        for (uint32_t i = 0; i < num_elems; i++) {
+        for (size_t i = 0; i < num_elems; i++) {
             cfish_Obj *val = CFISH_Vec_Fetch(self, i);
             if (val == NULL) {
                 continue;
             }
             else {
                 // Recurse for each value.
-                SV *const val_sv = (SV*)CFISH_Obj_To_Host(val);
+                SV *const val_sv = (SV*)CFISH_Obj_To_Host(val, cache);
                 av_store(perl_array, i, val_sv);
             }
         }
     }
 
+    if (cache == &new_cache && cache->seen) {
+        CFISH_PtrHash_Destroy(cache->seen);
+    }
+
     return newRV_noinc((SV*)perl_array);
 }
 
 /***************************** Clownfish::Hash ******************************/
 
 void*
-CFISH_Hash_To_Host_IMP(cfish_Hash *self) {
+CFISH_Hash_To_Host_IMP(cfish_Hash *self, void *vcache) {
     dTHX;
+    cfish_ConversionCache *cache = (cfish_ConversionCache*)vcache;
+    cfish_ConversionCache  new_cache;
+
+    if (cache) {
+        // Lookup Hash in conversion cache.
+        if ((cfish_Obj*)self == cache->root_obj) {
+            return newRV_inc(cache->root_sv);
+        }
+        if (cache->seen) {
+            void *cached_hv = CFISH_PtrHash_Fetch(cache->seen, self);
+            if (cached_hv) {
+                return newRV_inc((SV*)cached_hv);
+            }
+        }
+    }
+
     HV *perl_hash = newHV();
+
+    if (!cache) {
+        // Set up conversion cache.
+        cache = &new_cache;
+        cache->root_obj = (cfish_Obj*)self;
+        cache->root_sv  = (SV*)perl_hash;
+        cache->seen     = NULL;
+    }
+    else {
+        if (!cache->seen) {
+            // Create PtrHash lazily.
+            cache->seen = cfish_PtrHash_new(0);
+        }
+        CFISH_PtrHash_Store(cache->seen, self, perl_hash);
+    }
+
     cfish_HashIterator *iter = cfish_HashIter_new(self);
 
     // Iterate over key-value pairs.
@@ -927,13 +1129,19 @@ CFISH_Hash_To_Host_IMP(cfish_Hash *self) {
 
         // Recurse for each value.
         cfish_Obj *val    = CFISH_HashIter_Get_Value(iter);
-        SV        *val_sv = XSBind_cfish_to_perl(aTHX_ val);
+        SV        *val_sv = val
+                            ? (SV*)CFISH_Obj_To_Host(val, cache)
+                            : newSV(0);
 
         // Using a negative `klen` argument to signal UTF-8 is undocumented
         // in older Perl versions but works since 5.8.0.
         (void)hv_store(perl_hash, key_ptr, -key_size, val_sv, 0);
     }
 
+    if (cache == &new_cache && cache->seen) {
+        CFISH_PtrHash_Destroy(cache->seen);
+    }
+
     CFISH_DECREF(iter);
     return newRV_noinc((SV*)perl_hash);
 }
@@ -941,13 +1149,15 @@ CFISH_Hash_To_Host_IMP(cfish_Hash *self) {
 /****************************** Clownfish::Num ******************************/
 
 void*
-CFISH_Float_To_Host_IMP(cfish_Float *self) {
+CFISH_Float_To_Host_IMP(cfish_Float *self, void *vcache) {
+    CFISH_UNUSED_VAR(vcache);
     dTHX;
     return newSVnv(self->value);
 }
 
 void*
-CFISH_Int_To_Host_IMP(cfish_Integer *self) {
+CFISH_Int_To_Host_IMP(cfish_Integer *self, void *vcache) {
+    CFISH_UNUSED_VAR(vcache);
     dTHX;
     SV *sv = NULL;
 
@@ -962,7 +1172,8 @@ CFISH_Int_To_Host_IMP(cfish_Integer *self) {
 }
 
 void*
-CFISH_Bool_To_Host_IMP(cfish_Boolean *self) {
+CFISH_Bool_To_Host_IMP(cfish_Boolean *self, void *vcache) {
+    CFISH_UNUSED_VAR(vcache);
     dTHX;
     return newSViv((IV)self->value);
 }
